"""A climate model workflow management tool.

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
from . import _version
__version__ = _version.get_versions()['version']
